@@ -2,11 +2,12 @@ import dataclasses
 import functools as ft
 import inspect
 import warnings
-from typing import Any, Callable, Dict, Hashable, Optional, overload, Union
+from collections.abc import Callable, Hashable
+from typing import Any, Literal, overload
 
 import jax
-import jax.interpreters.batching as batching
-import jax.interpreters.pxla as pxla
+import jax._src.traceback_util as traceback_util
+import jax.core
 import jax.numpy as jnp
 import jax.tree_util as jtu
 import numpy as np
@@ -14,19 +15,22 @@ from jaxtyping import PyTree
 
 from ._compile_utils import (
     compile_cache,
-    get_fun_names,
     hashable_combine,
     hashable_partition,
+    Lowered,
 )
 from ._custom_types import sentinel
 from ._deprecate import deprecated_0_10
 from ._doc_utils import doc_remove_args
 from ._filters import combine, filter, is_array, is_array_like, partition
-from ._module import Module, module_update_wrapper, Static
+from ._module import Module, module_update_wrapper, Partial, Static
 
 
-ResolvedAxisSpec = Optional[int]
-AxisSpec = Union[ResolvedAxisSpec, Callable[[Any], ResolvedAxisSpec]]
+traceback_util.register_exclusion(__file__)
+
+
+ResolvedAxisSpec = None | int
+AxisSpec = ResolvedAxisSpec | Callable[[Any], ResolvedAxisSpec]
 
 
 def _is_none(x: Any) -> bool:
@@ -68,42 +72,12 @@ class if_array:
 
     axis: int
 
-    def __call__(self, x: Any) -> Optional[int]:
+    def __call__(self, x: Any) -> int | None:
         return self.axis if is_array(x) else None
 
 
-@dataclasses.dataclass(frozen=True)  # not a pytree
-class if_mapped:
-    """Returns a callable that returns the specified integer if evaluated on a mapped
-    output from a vmap'd or pmap'd function. Otherwise, it returns `None`.
-
-    !!! Example
-
-        ```python
-        fn = if_array(1)
-        # Evaluate on an array, return the integer.
-        fn(jax.numpy.array([0, 1, 2]))  # 1
-        # Evaluate on not-an-array, return None.
-        fn(True)  # None
-        ```
-    """
-
-    axis: int
-
-    def __call__(self, x: Any) -> Optional[int]:
-        if isinstance(x, batching.BatchTracer):
-            if x.batch_dim is batching.not_mapped:
-                return None
-            else:
-                return self.axis
-        elif isinstance(x, pxla.MapTracer):
-            return self.axis
-        else:
-            return None
-
-
-def _swapaxes(array, axis):
-    return jnp.swapaxes(array, 0, axis)
+def _moveaxis(array, axis):
+    return jnp.moveaxis(array, 0, axis)
 
 
 def _named_in_axes(fun, in_axes, args):
@@ -113,7 +87,7 @@ def _named_in_axes(fun, in_axes, args):
         default = if_array(0)
         params = inspect.signature(fun).parameters
         # We may have that len(args) < len(params) due to default arguments.
-        # Truncate to considering just the argments that have been passed.
+        # Truncate to considering just the arguments that have been passed.
         #
         # (If len(args) > len(params) then we'll get the usual error later when
         # attempting to call with the wrong number of arguments.)
@@ -143,9 +117,13 @@ class _VmapWrapper(Module):
     _fun: Callable
     _in_axes: PyTree[AxisSpec]
     _out_axes: PyTree[AxisSpec]
-    _axis_name: Optional[Hashable]
-    _axis_size: Optional[int]
-    _vmapkwargs: Dict[str, Any]
+    _axis_name: Hashable | None
+    _axis_size: int | None
+    _vmapkwargs: dict[str, Any]
+
+    @property
+    def __wrapped__(self):
+        return self._fun
 
     def __call__(self, /, *args, **kwargs):
         if len(kwargs) != 0:
@@ -176,36 +154,40 @@ class _VmapWrapper(Module):
             _out = self._fun(*_args)
             _out_axes = _resolve_axes(_out, self._out_axes)
             _none_axes = jtu.tree_map(_is_none, _out_axes, is_leaf=_is_none)
-            _nonvmapd, _vmapd = partition(_out, _none_axes)
-            return _vmapd, Static((_nonvmapd, _out_axes))
+            _nonvmapd, _vmapd = partition(_out, _none_axes, is_leaf=_is_none)
+            _nonvmapd_arr, _nonvmapd_static = partition(_nonvmapd, is_array)
+            return _vmapd, _nonvmapd_arr, Static((_nonvmapd_static, _out_axes))
 
         if len(jtu.tree_leaves(in_axes)) == 0 and self._axis_size is None:
-            vmapd, static = _fun_wrapper(dynamic_args)
+            vmapd, nonvmapd_arr, static = _fun_wrapper(dynamic_args)
             if len(jtu.tree_leaves(vmapd)) != 0:
                 raise ValueError(
                     "Cannot resolve batch dimension. Non-`None` `out_axes` requires "
                     "either `in_axes` or `axis_size` to be not `None`."
                 )
         else:
-            vmapd, static = jax.vmap(
+            vmapd, nonvmapd_arr, static = jax.vmap(
                 _fun_wrapper,
                 in_axes=(in_axes,),
-                out_axes=(0, None),
+                out_axes=(0, None, None),
                 axis_name=self._axis_name,
                 axis_size=self._axis_size,
                 **self._vmapkwargs,
             )(dynamic_args)
-        nonvmapd, out_axes = static.value
+
+        nonvmapd_static, out_axes = static.value
+        nonvmapd = combine(nonvmapd_arr, nonvmapd_static)
 
         assert jtu.tree_structure(vmapd) == jtu.tree_structure(out_axes)
-        vmapd = jtu.tree_map(_swapaxes, vmapd, out_axes)
+        vmapd = jtu.tree_map(_moveaxis, vmapd, out_axes)
 
         return combine(vmapd, nonvmapd)
 
     def __get__(self, instance, owner):
+        del owner
         if instance is None:
             return self
-        return jtu.Partial(self, instance)
+        return Partial(self, instance)
 
 
 @overload
@@ -214,9 +196,8 @@ def filter_vmap(
     in_axes: PyTree[AxisSpec] = if_array(0),
     out_axes: PyTree[AxisSpec] = if_array(0),
     axis_name: Hashable = None,
-    axis_size: Optional[int] = None,
-) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
-    ...
+    axis_size: int | None = None,
+) -> Callable[[Callable[..., Any]], Callable[..., Any]]: ...
 
 
 @overload
@@ -226,9 +207,8 @@ def filter_vmap(
     in_axes: PyTree[AxisSpec] = if_array(0),
     out_axes: PyTree[AxisSpec] = if_array(0),
     axis_name: Hashable = None,
-    axis_size: Optional[int] = None,
-) -> Callable[..., Any]:
-    ...
+    axis_size: int | None = None,
+) -> Callable[..., Any]: ...
 
 
 @doc_remove_args("vmapkwargs")
@@ -238,7 +218,7 @@ def filter_vmap(
     in_axes: PyTree[AxisSpec] = if_array(0),
     out_axes: PyTree[AxisSpec] = if_array(0),
     axis_name: Hashable = None,
-    axis_size: Optional[int] = None,
+    axis_size: int | None = None,
     **vmapkwargs,
 ):
     """Vectorises a function. By default, all JAX/NumPy arrays are vectorised down their
@@ -257,7 +237,7 @@ def filter_vmap(
         It should be a PyTree of `None`, `int`, or callables `Leaf -> Union[None, int]`.
         Its tree structure should either be:
         1. a prefix of the input tuple of `args`.
-        2. a dictionary, in which case the named argments will use the specified indices
+        2. a dictionary, in which case the named arguments use the specified indices
             to vectorise over, and all other arguments will have the default
             `eqx.if_array(0)`.
     - `out_axes` indicates which axis of the output arrays the mapped axis should appear
@@ -377,15 +357,13 @@ def filter_vmap(
         _axis_size=axis_size,
         _vmapkwargs=vmapkwargs,
     )
-    return module_update_wrapper(vmap_wrapper, fun)
+    return module_update_wrapper(vmap_wrapper)
 
 
 @compile_cache
 def _filter_pmap_cache(
-    fun_names, leaves, treedef, in_axes, axis_name, axis_size, pmapkwargs
+    fun_names, static, struct, in_axes, axis_name, axis_size, pmapkwargs
 ):
-    static, struct = jtu.tree_unflatten(treedef, leaves)
-
     def fun_abstract(_dynamic):
         fun, args, _, _ = combine(_dynamic, static)
         out = fun(*args)
@@ -398,7 +376,7 @@ def _filter_pmap_cache(
     max_out_size = jtu.tree_reduce(lambda x, y: max(x, y.ndim), struct_out, 0)
     del fun_abstract, struct, struct_out
 
-    def _check_map_out_axis(x: Optional[int]):
+    def _check_map_out_axis(x: int | None):
         if isinstance(x, int):
             if x < -max_out_size or x >= max_out_size:
                 raise ValueError(
@@ -417,10 +395,10 @@ def _filter_pmap_cache(
         jtu.tree_map(_check_map_out_axis, _out_axes)
         _pmapd = []
         for i in range(-max_out_size, max_out_size):
-            _i_axes = jtu.tree_map(lambda a: a == i, _out_axes)
-            _pmapd.append(filter(_out, _i_axes))
+            _i_axes = jtu.tree_map(lambda a: a == i, _out_axes, is_leaf=_is_none)
+            _pmapd.append(filter(_out, _i_axes, is_leaf=_is_none))
         _none_axes = jtu.tree_map(_is_none, _out_axes, is_leaf=_is_none)
-        _nonpmapd = filter(_out, _none_axes)
+        _nonpmapd = filter(_out, _none_axes, is_leaf=_is_none)
         _dynamic_nonpmapd, _static_nonpmapd = hashable_partition(_nonpmapd, is_array)
         return _pmapd, _dynamic_nonpmapd, Static(_static_nonpmapd)
 
@@ -440,29 +418,49 @@ def _filter_pmap_cache(
     )
 
 
+def _common_preprocess(axis_size, kwargs):
+    if len(kwargs) != 0:
+        raise RuntimeError(
+            "keyword arguments cannot be used with functions wrapped with `filter_pmap`"
+        )
+    if axis_size is None:
+        return 0  # hashable non-array object
+    else:
+        # Work around JAX issue #9252
+        return np.broadcast_to(0, axis_size)
+
+
+def _preprocess(info, args, kwargs):
+    fun, out_axes, axis_size = info
+    maybe_dummy = _common_preprocess(axis_size, kwargs)
+    del kwargs
+    dynamic = filter((fun, args, maybe_dummy, out_axes), is_array)
+    return (dynamic,)
+
+
+def _postprocess(out):
+    (pmapd, dynamic_nonpmapd, static_nonpmapd) = out
+    nonpmapd = hashable_combine(dynamic_nonpmapd, static_nonpmapd.value)
+    return combine(*pmapd, nonpmapd)
+
+
 class _PmapWrapper(Module):
     _fun: Callable
     _in_axes: PyTree[AxisSpec]
     _out_axes: PyTree[AxisSpec]
-    _axis_name: Optional[Hashable]
-    _axis_size: Optional[int]
+    _axis_name: Hashable | None
+    _axis_size: int | None
     _filter_warning: bool
-    _pmapkwargs: Dict[str, Any]
+    _pmapkwargs: dict[str, Any]
+
+    @property
+    def __wrapped__(self):
+        return self._fun
 
     def _call(self, is_lower, args, kwargs):
-        if len(kwargs) != 0:
-            raise RuntimeError(
-                "keyword arguments cannot be used with functions wrapped with "
-                "`filter_pmap`"
-            )
+        maybe_dummy = _common_preprocess(self._axis_size, kwargs)
         del kwargs
 
-        if self._axis_size is None:
-            maybe_dummy = 0  # hashable non-array object
-        else:
-            # Work around JAX issue #9252
-            maybe_dummy = np.broadcast_to(0, self._axis_size)
-
         in_axes = _named_in_axes(self._fun, self._in_axes, args)
         in_axes = _resolve_axes(args, in_axes)
         in_axes = (None, in_axes, 0, None)
@@ -471,13 +469,11 @@ class _PmapWrapper(Module):
             (self._fun, args, maybe_dummy, self._out_axes), is_array
         )
         struct = jtu.tree_map(lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype), dynamic)
-        leaves, treedef = jtu.tree_flatten((static, struct))
-        leaves = tuple(leaves)
 
         cached = _filter_pmap_cache(
-            get_fun_names(self._fun),
-            leaves,
-            treedef,
+            self._fun,
+            static,
+            struct,
             in_axes,
             self._axis_name,
             self._axis_size,
@@ -485,29 +481,34 @@ class _PmapWrapper(Module):
         )
 
         if is_lower:
-            return cached.lower(dynamic)
+            return Lowered(
+                cached.lower(dynamic),
+                (self._fun, self._out_axes, self._axis_size),
+                _preprocess,  # pyright: ignore
+                _postprocess,  # pyright: ignore
+            )
         else:
-            (pmapd, dynamic_nonpmapd, static_nonpmapd) = cached(dynamic)
-            nonpmapd = hashable_combine(dynamic_nonpmapd, static_nonpmapd.value)
-            return combine(*pmapd, nonpmapd)
+            if self._filter_warning is True:
+                with warnings.catch_warnings():
+                    warnings.filterwarnings(
+                        "ignore", message="Some donated buffers were not usable*"
+                    )
+                    out = cached(dynamic)
+            else:
+                out = cached(dynamic)
+            return _postprocess(out)
 
     def __call__(self, /, *args, **kwargs):
-        if self._filter_warning is True:
-            with warnings.catch_warnings():
-                warnings.filterwarnings(
-                    "ignore", message="Some donated buffers were not usable*"
-                )
-                return self._call(False, args, kwargs)
-        else:
-            return self._call(False, args, kwargs)
+        return self._call(False, args, kwargs)
 
-    def lower(self, /, *args, **kwargs):
+    def lower(self, /, *args, **kwargs) -> Lowered:
         return self._call(True, args, kwargs)
 
     def __get__(self, instance, owner):
+        del owner
         if instance is None:
             return self
-        return jtu.Partial(self, instance)
+        return Partial(self, instance)
 
 
 # Deliberately using `Callable[..., Any]` as `filter_pmap` does change the input and
@@ -519,10 +520,9 @@ def filter_pmap(
     in_axes: PyTree[AxisSpec] = if_array(0),
     out_axes: PyTree[AxisSpec] = if_array(0),
     axis_name: Hashable = None,
-    axis_size: Optional[int] = None,
-    donate: str = "none",
-) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
-    ...
+    axis_size: int | None = None,
+    donate: Literal["all", "warn", "none"] = "none",
+) -> Callable[[Callable[..., Any]], Callable[..., Any]]: ...
 
 
 @overload
@@ -532,10 +532,9 @@ def filter_pmap(
     in_axes: PyTree[AxisSpec] = if_array(0),
     out_axes: PyTree[AxisSpec] = if_array(0),
     axis_name: Hashable = None,
-    axis_size: Optional[int] = None,
-    donate: str = "none",
-) -> Callable[..., Any]:
-    ...
+    axis_size: int | None = None,
+    donate: Literal["all", "warn", "none"] = "none",
+) -> Callable[..., Any]: ...
 
 
 @doc_remove_args("pmapkwargs")
@@ -545,11 +544,19 @@ def filter_pmap(
     in_axes: PyTree[AxisSpec] = if_array(0),
     out_axes: PyTree[AxisSpec] = if_array(0),
     axis_name: Hashable = None,
-    axis_size: Optional[int] = None,
-    donate: str = "none",
+    axis_size: int | None = None,
+    donate: Literal["all", "warn", "none"] = "none",
     **pmapkwargs,
 ):
-    """Parallelises a function. By default, all JAX/NumPy arrays are parallelised down
+    """
+    !!! warning
+
+        JAX has now added more powerful parallelism APIs directly to the JIT interface.
+        As such, using [`equinox.filter_jit`][] with sharded inputs is now recommended
+        over `filter_pmap`. See also the
+        [parallelism example](../examples/parallelism.ipynb).
+
+    Parallelises a function. By default, all JAX/NumPy arrays are parallelised down
     their leading axis (i.e. axis index 0), and all other types are broadcast.
 
     `jax.pmap`, and thus `equinox.filter_pmap`, also compiles their function in the same
@@ -569,7 +576,7 @@ def filter_pmap(
         It should be a PyTree of `None`, `int`, or callables `Leaf -> Union[None, int]`.
         Its tree structure should either be:
         1. a prefix of the input tuple of `args`.
-        2. a dictionary, in which case the named argments will use the specified indices
+        2. a dictionary, in which case the named arguments use the specified indices
             to parallelise over, and all other arguments will have the default
             `eqx.if_array(0)`.
     - `out_axes` indicates which axis of the output arrays the mapped axis should appear
@@ -583,10 +590,10 @@ def filter_pmap(
         it can be deduced by looking at the argument shapes.
     - `donate` indicates whether the buffers of JAX arrays are donated or not, it
         should either be:
-        - `'all'`: the default, donate all arrays and suppress all warnings about
+        - `'all'`: donate all arrays and suppress all warnings about
             unused buffers;
         - `'warn'`: as above, but don't suppress unused buffer warnings;
-        - `'none'`: disables buffer donation.
+        - `'none'`: the default, disables buffer donation.
 
     **Returns:**
 
@@ -646,11 +653,18 @@ def filter_pmap(
             "'donate_argnums'"
         )
 
-    if donate not in {"arrays", "warn", "none"}:
-        raise ValueError(
-            "`filter_jit(..., donate=...)` must be one of 'arrays', 'warn', or 'none'"
+    if donate == "arrays":
+        warnings.warn(
+            "The `donate='arrays'` option to `filter_pmap` has been renamed to "
+            "`donate='all'`",
+            DeprecationWarning,
         )
-    filter_warning = True if donate == "arrays" else False
+        donate = "all"
+    if donate not in {"all", "warn", "none"}:
+        raise ValueError(
+            "`filter_jit(..., donate=...)` must be one of 'all', 'warn', or 'none'"
+        )
+    filter_warning = True if donate == "all" else False
     if donate != "none":
         pmapkwargs["donate_argnums"] = (0,)
 
@@ -663,4 +677,4 @@ def filter_pmap(
         _filter_warning=filter_warning,
         _pmapkwargs=pmapkwargs,
     )
-    return module_update_wrapper(pmap_wrapper, fun)
+    return module_update_wrapper(pmap_wrapper)
