@@ -1,9 +1,9 @@
-from typing import Optional, Union
-
+import jax
 import jax.numpy as jnp
 from jaxtyping import Array
 
 from .._module import Module
+from ._misc import named_scope
 
 
 class PReLU(Module):
@@ -17,7 +17,7 @@ class PReLU(Module):
 
     def __init__(
         self,
-        init_alpha: Optional[Union[float, Array]] = 0.25,
+        init_alpha: float | Array | None = 0.25,
     ):
         r"""**Arguments:**
 
@@ -29,6 +29,7 @@ class PReLU(Module):
 
         self.negative_slope = jnp.asarray(init_alpha)
 
+    @named_scope("eqx.nn.PReLU")
     def __call__(self, x: Array) -> Array:
         r"""**Arguments:**
 
@@ -38,4 +39,5 @@ class PReLU(Module):
 
         A JAX array of the same shape as the input.
         """
-        return jnp.where(x >= 0, x, self.negative_slope * x)
+        with jax.numpy_dtype_promotion("standard"), jax.numpy_rank_promotion("allow"):
+            return jnp.where(x >= 0, x, self.negative_slope * x)
