@@ -1,38 +1,36 @@
-from typing import Callable, Optional, Sequence, Tuple, Union
+import math
+from collections.abc import Callable, Sequence
 
 import jax
 import jax.lax as lax
 import jax.numpy as jnp
 import jax.random
-import numpy as np
-from jaxtyping import Array
+from jaxtyping import Array, PRNGKeyArray
 
-from .._custom_types import PRNGKey
-from .._module import Module, static_field
-from ._misc import all_sequences
+from .._module import field, Module
+from ._misc import all_sequences, named_scope
 
 
 class Pool(Module):
     """General N-dimensional downsampling over a sliding window."""
 
-    init: Union[int, float, Array]
+    init: int | float | Array
     operation: Callable[[Array, Array], Array]
-    num_spatial_dims: int = static_field()
-    kernel_size: Tuple[int, ...] = static_field()
-    stride: Tuple[int, ...] = static_field()
-    padding: Tuple[Tuple[int, int], ...] = static_field()
-    use_ceil: bool = static_field()
+    num_spatial_dims: int = field(static=True)
+    kernel_size: tuple[int, ...] = field(static=True)
+    stride: tuple[int, ...] = field(static=True)
+    padding: tuple[tuple[int, int], ...] = field(static=True)
+    use_ceil: bool = field(static=True)
 
     def __init__(
         self,
-        init: Union[int, float, Array],
+        init: int | float | Array,
         operation: Callable[[Array, Array], Array],
         num_spatial_dims: int,
-        kernel_size: Union[int, Sequence[int]],
-        stride: Union[int, Sequence[int]] = 1,
-        padding: Union[int, Sequence[int], Sequence[Tuple[int, int]]] = 0,
+        kernel_size: int | Sequence[int],
+        stride: int | Sequence[int] = 1,
+        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
         use_ceil: bool = False,
-        **kwargs,
     ):
         """**Arguments:**
 
@@ -54,7 +52,6 @@ class Pool(Module):
             [https://www.tensorflow.org/xla/operation_semantics#reducewindow](https://www.tensorflow.org/xla/operation_semantics#reducewindow)
             and [https://github.com/google/jax/issues/7718](https://github.com/google/jax/issues/7718).
         """  # noqa: E501
-        super().__init__(**kwargs)
 
         self.operation = operation
         self.init = init
@@ -85,7 +82,7 @@ class Pool(Module):
             self.padding = tuple((padding, padding) for _ in range(num_spatial_dims))
         elif isinstance(padding, Sequence) and len(padding) == num_spatial_dims:
             if all_sequences(padding):
-                self.padding = tuple(padding)
+                self.padding = tuple(padding)  # pyright: ignore
             else:
                 self.padding = tuple((p, p) for p in padding)
         else:
@@ -116,7 +113,8 @@ class Pool(Module):
                     f"{kernel_size}."
                 )
 
-    def __call__(self, x: Array, *, key: Optional[PRNGKey] = None) -> Array:
+    @named_scope("eqx.nn.Pool")
+    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
         """**Arguments:**
 
         - `x`: The input. Should be a JAX array of shape
@@ -161,11 +159,10 @@ class AvgPool1d(Pool):
 
     def __init__(
         self,
-        kernel_size: Union[int, Sequence[int]],
-        stride: Union[int, Sequence[int]] = 1,
-        padding: Union[int, Sequence[int], Sequence[Tuple[int, int]]] = 0,
+        kernel_size: int | Sequence[int],
+        stride: int | Sequence[int] = 1,
+        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
         use_ceil: bool = False,
-        **kwargs,
     ):
         """**Arguments:**
 
@@ -186,10 +183,10 @@ class AvgPool1d(Pool):
             stride=stride,
             padding=padding,
             use_ceil=use_ceil,
-            **kwargs,
         )
 
-    def __call__(self, x: Array, *, key: Optional[PRNGKey] = None) -> Array:
+    @named_scope("eqx.nn.AvgPool1d")
+    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
         """**Arguments:**
 
         - `x`: The input. Should be a JAX array of shape `(channels, dim)`.
@@ -201,7 +198,7 @@ class AvgPool1d(Pool):
         A JAX array of shape `(channels, new_dim)`.
         """
 
-        return super().__call__(x) / np.prod(self.kernel_size)
+        return super().__call__(x) / math.prod(self.kernel_size)
 
 
 class MaxPool1d(Pool):
@@ -209,11 +206,10 @@ class MaxPool1d(Pool):
 
     def __init__(
         self,
-        kernel_size: Union[int, Sequence[int]],
-        stride: Union[int, Sequence[int]] = 1,
-        padding: Union[int, Sequence[int], Sequence[Tuple[int, int]]] = 0,
+        kernel_size: int | Sequence[int],
+        stride: int | Sequence[int] = 1,
+        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
         use_ceil: bool = False,
-        **kwargs,
     ):
         """**Arguments:**
 
@@ -234,11 +230,11 @@ class MaxPool1d(Pool):
             stride=stride,
             padding=padding,
             use_ceil=use_ceil,
-            **kwargs,
         )
 
     # Redefined to get them in the right order in docs
-    def __call__(self, x: Array, *, key: Optional[PRNGKey] = None) -> Array:
+    @named_scope("eqx.nn.MaxPool1d")
+    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
         """**Arguments:**
 
         - `x`: The input. Should be a JAX array of shape `(channels, dim)`.
@@ -258,11 +254,10 @@ class AvgPool2d(Pool):
 
     def __init__(
         self,
-        kernel_size: Union[int, Sequence[int]],
-        stride: Union[int, Sequence[int]] = 1,
-        padding: Union[int, Sequence[int], Sequence[Tuple[int, int]]] = 0,
+        kernel_size: int | Sequence[int],
+        stride: int | Sequence[int] = 1,
+        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
         use_ceil: bool = False,
-        **kwargs,
     ):
         """**Arguments:**
 
@@ -283,10 +278,10 @@ class AvgPool2d(Pool):
             stride=stride,
             padding=padding,
             use_ceil=use_ceil,
-            **kwargs,
         )
 
-    def __call__(self, x: Array, *, key: Optional[PRNGKey] = None) -> Array:
+    @named_scope("eqx.nn.AvgPool2d")
+    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
         """**Arguments:**
 
         - `x`: The input. Should be a JAX array of shape `(channels, dim_1, dim_2)`.
@@ -298,7 +293,7 @@ class AvgPool2d(Pool):
         A JAX array of shape `(channels, new_dim_1, new_dim_2)`.
         """
 
-        return super().__call__(x) / np.prod(self.kernel_size)
+        return super().__call__(x) / math.prod(self.kernel_size)
 
 
 class MaxPool2d(Pool):
@@ -306,11 +301,10 @@ class MaxPool2d(Pool):
 
     def __init__(
         self,
-        kernel_size: Union[int, Sequence[int]],
-        stride: Union[int, Sequence[int]] = 1,
-        padding: Union[int, Sequence[int], Sequence[Tuple[int, int]]] = 0,
+        kernel_size: int | Sequence[int],
+        stride: int | Sequence[int] = 1,
+        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
         use_ceil: bool = False,
-        **kwargs,
     ):
         """**Arguments:**
 
@@ -331,11 +325,11 @@ class MaxPool2d(Pool):
             stride=stride,
             padding=padding,
             use_ceil=use_ceil,
-            **kwargs,
         )
 
     # Redefined to get them in the right order in docs
-    def __call__(self, x: Array, *, key: Optional[PRNGKey] = None) -> Array:
+    @named_scope("eqx.nn.MaxPool2d")
+    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
         """**Arguments:**
 
         - `x`: The input. Should be a JAX array of shape `(channels, dim_1, dim_2)`.
@@ -355,11 +349,10 @@ class AvgPool3d(Pool):
 
     def __init__(
         self,
-        kernel_size: Union[int, Sequence[int]],
-        stride: Union[int, Sequence[int]] = 1,
-        padding: Union[int, Sequence[int], Sequence[Tuple[int, int]]] = 0,
+        kernel_size: int | Sequence[int],
+        stride: int | Sequence[int] = 1,
+        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
         use_ceil: bool = False,
-        **kwargs,
     ):
         """**Arguments:**
 
@@ -380,10 +373,10 @@ class AvgPool3d(Pool):
             stride=stride,
             padding=padding,
             use_ceil=use_ceil,
-            **kwargs,
         )
 
-    def __call__(self, x: Array, *, key: Optional[PRNGKey] = None) -> Array:
+    @named_scope("eqx.nn.AvgPool3d")
+    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
         """**Arguments:**
 
         - `x`: The input. Should be a JAX array of shape
@@ -396,7 +389,7 @@ class AvgPool3d(Pool):
         A JAX array of shape `(channels, new_dim_1, new_dim_2, new_dim_3)`.
         """
 
-        return super().__call__(x) / np.prod(self.kernel_size)
+        return super().__call__(x) / math.prod(self.kernel_size)
 
 
 class MaxPool3d(Pool):
@@ -404,11 +397,10 @@ class MaxPool3d(Pool):
 
     def __init__(
         self,
-        kernel_size: Union[int, Sequence[int]],
-        stride: Union[int, Sequence[int]] = 1,
-        padding: Union[int, Sequence[int], Sequence[Tuple[int, int]]] = 0,
+        kernel_size: int | Sequence[int],
+        stride: int | Sequence[int] = 1,
+        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
         use_ceil: bool = False,
-        **kwargs,
     ):
         """**Arguments:**
 
@@ -429,10 +421,10 @@ class MaxPool3d(Pool):
             stride=stride,
             padding=padding,
             use_ceil=use_ceil,
-            **kwargs,
         )
 
-    def __call__(self, x: Array, *, key: Optional[PRNGKey] = None) -> Array:
+    @named_scope("eqx.nn.MaxPool3d")
+    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
         """**Arguments:**
 
         - `x`: The input. Should be a JAX array of shape
@@ -481,15 +473,14 @@ def _adaptive_pool1d(
 class AdaptivePool(Module):
     """General N dimensional adaptive downsampling to a target shape."""
 
-    target_shape: Sequence[int] = static_field()
+    target_shape: Sequence[int] = field(static=True)
     operation: Callable[[Array], Array]
 
     def __init__(
         self,
-        target_shape: Union[int, Sequence[int]],
+        target_shape: int | Sequence[int],
         num_spatial_dims: int,
         operation: Callable,
-        **kwargs,
     ):
         """**Arguments:**
 
@@ -497,7 +488,6 @@ class AdaptivePool(Module):
         - `num_spatial_dims`: The number of spatial dimensions.
         - `operation`: The operation applied for downsample.
         """
-        super().__init__(**kwargs)
         self.operation = operation
         if isinstance(target_shape, int):
             self.target_shape = (target_shape,) * num_spatial_dims
@@ -511,7 +501,8 @@ class AdaptivePool(Module):
                 f"{num_spatial_dims} containing ints."
             )
 
-    def __call__(self, x: Array, *, key: Optional[PRNGKey] = None) -> Array:
+    @named_scope("eqx.nn.AdaptivePool")
+    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
         """**Arguments:**
 
         - `x`: The input. Should be a JAX array of shape
@@ -526,7 +517,7 @@ class AdaptivePool(Module):
         if x.ndim - 1 != len(self.target_shape):
             raise ValueError(
                 f"Expected input with {len(self.target_shape)} dimensions, "
-                f"received {x.ndim-1} instead."
+                f"received {x.ndim - 1} instead."
             )
         for i in range(1, x.ndim):
             op = jax.vmap(
@@ -543,64 +534,64 @@ class AdaptivePool(Module):
 class AdaptiveAvgPool1d(AdaptivePool):
     """Adaptive one-dimensional downsampling using average for the target shape."""
 
-    def __init__(self, target_shape: Union[int, Sequence[int]], **kwargs):
+    def __init__(self, target_shape: int | Sequence[int]):
         """**Arguments:**
 
         - `target_shape`: The target output shape.
         """
-        super().__init__(target_shape, num_spatial_dims=1, operation=jnp.mean, **kwargs)
+        super().__init__(target_shape, num_spatial_dims=1, operation=jnp.mean)
 
 
 class AdaptiveAvgPool2d(AdaptivePool):
     """Adaptive two-dimensional downsampling using average for the target shape."""
 
-    def __init__(self, target_shape: Union[int, Sequence[int]], **kwargs):
+    def __init__(self, target_shape: int | Sequence[int]):
         """**Arguments:**
 
         - `target_shape`: The target output shape.
         """
-        super().__init__(target_shape, num_spatial_dims=2, operation=jnp.mean, **kwargs)
+        super().__init__(target_shape, num_spatial_dims=2, operation=jnp.mean)
 
 
 class AdaptiveAvgPool3d(AdaptivePool):
     """Adaptive three-dimensional downsampling using average for the target shape."""
 
-    def __init__(self, target_shape: Union[int, Sequence[int]], **kwargs):
+    def __init__(self, target_shape: int | Sequence[int]):
         """**Arguments:**
 
         - `target_shape`: The target output shape.
         """
-        super().__init__(target_shape, num_spatial_dims=3, operation=jnp.mean, **kwargs)
+        super().__init__(target_shape, num_spatial_dims=3, operation=jnp.mean)
 
 
 class AdaptiveMaxPool1d(AdaptivePool):
     """Adaptive one-dimensional downsampling using maximum for the target shape."""
 
-    def __init__(self, target_shape: Union[int, Sequence[int]], **kwargs):
+    def __init__(self, target_shape: int | Sequence[int]):
         """**Arguments:**
 
         - `target_shape`: The target output shape.
         """
-        super().__init__(target_shape, num_spatial_dims=1, operation=jnp.max, **kwargs)
+        super().__init__(target_shape, num_spatial_dims=1, operation=jnp.max)
 
 
 class AdaptiveMaxPool2d(AdaptivePool):
     """Adaptive two-dimensional downsampling using maximum for the target shape."""
 
-    def __init__(self, target_shape: Union[int, Sequence[int]], **kwargs):
+    def __init__(self, target_shape: int | Sequence[int]):
         """**Arguments:**
 
         - `target_shape`: The target output shape.
         """
-        super().__init__(target_shape, num_spatial_dims=2, operation=jnp.max, **kwargs)
+        super().__init__(target_shape, num_spatial_dims=2, operation=jnp.max)
 
 
 class AdaptiveMaxPool3d(AdaptivePool):
     """Adaptive three-dimensional downsampling using maximum for the target shape."""
 
-    def __init__(self, target_shape: Union[int, Sequence[int]], **kwargs):
+    def __init__(self, target_shape: int | Sequence[int]):
         """**Arguments:**
 
         - `target_shape`: The target output shape.
         """
-        super().__init__(target_shape, num_spatial_dims=3, operation=jnp.max, **kwargs)
+        super().__init__(target_shape, num_spatial_dims=3, operation=jnp.max)
